def endpoint(service: str = '') -> str:
    """ Obtain the final URL to the service dynamically from a request.

    :param request: The request to obtain the URL.
    :param service: The path to the services to join with the request base.
    :return: A string with the URL.
    """
    base_url = '<script>document.write(window.location.href.replace(/\\/$/, ""));</script>'
    return base_url + service if service.startswith('/') else base_url + '/' + service



import re

URL_PATTERN = r'[A-Za-z0-9]+://[A-Za-z0-9%-_]+(/[A-Za-z0-9%-_])*(#|\\?)[A-Za-z0-9%-_&=]*'


def remove_urls(text: str) -> str:
    """ Remove any url in the text.

    :param text: The text to remove urls.
    :return: The same text but without urls.
    """
    return re.sub(URL_PATTERN, '', text)


def clean_text(text: str, lower: bool = True, url: bool = True) -> str:
    """ Remove punctuation symbols, urls and convert to lower.

    :param text: The text to clean.
    :param lower: If you want to convert to lower.
    :param url: If you want to remove urls.
    :return: The cleaned text.
    """
    text = text.lower() if lower else text
    text = remove_urls(text) if url else text
    text = re.sub(r'[^A-Za-z0-9 -]', ' ', text)
    return re.sub(r'[\s]+', ' ', text).strip()



import logging
from typing import List, Tuple, Union

LOG_FORMAT = '%(asctime)s %(levelname)-8s %(message)s'
DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


def get_log_levels() -> List[Tuple[str, int]]:
    """ Obtain the defined log level names and its integer representation.
    :return: A list of tuples with the official level name and its integer value.
    """
    return [(k, v) for k, v in logging._nameToLevel.items()]


def get_log_level_names(low_case: bool = True) -> List[str]:
    """ Obtain a list with the level names.
    :param low_case: If convert the level names into lower case.
    :return: A list of level names.
    """
    return [k.lower() if low_case else k for k in logging._nameToLevel]


def get_log_level(name: str) -> int:
    """ Return the number representation of a level from its name.

    :param name: The level name.
    :return: Its integer value.
    """
    return logging._nameToLevel[name.upper()]


def config_log(level: Union[str, int], filename: str = None) -> None:
    """  Configure easily the logging using a default log and date format.

    :param level: The level name or number.
    :param filename: The file name to store the logs. By default, the standard error output is used.
    """
    level = level.upper() if isinstance(level, str) else level
    if filename:
        logging.basicConfig(level=level, filename=filename, format=LOG_FORMAT, datefmt=DATE_FORMAT)
    else:
        logging.basicConfig(level=level, format=LOG_FORMAT, datefmt=DATE_FORMAT)



import gzip
import json
import pickle
import tarfile
from os import makedirs
from tarfile import TarInfo
from os.path import basename, isdir, join, exists, splitext, dirname
from typing import List, Any

from typing.io import IO

# Import tqdm if it is installed, otherwise a dummy tqdm function is used.
try:
    from tqdm.auto import tqdm
except ModuleNotFoundError as e:
    def tqdm(obj: Any) -> Any:
        return obj

# Available tar compress methods
COMPRESS_METHODS = {'gz', 'bz2', 'xz'}


def create_tar(filename: str, *files: str, verbose: bool = False) -> None:
    """
    Create a tar file with a given list of files. If the filename has any of these extensions 'gz', 'bz2' or 'xz',
      then the tar file will be compressed with the specify method.
    :param filename: The name of the tar file.
    :param files: The list of file paths to include in the tar file.
    :param verbose: if True, show a bar progress.
    """
    compress_method = detect_compress_method(filename)
    with tarfile.open(filename, f'w:{compress_method}') as tar:
        for file in tqdm(files, desc='Creating tar file', disable=not verbose):
            tar.add(file, basename(file))


def detect_compress_method(filename: str) -> str:
    """ Detecting the compression method based on the extension of the filename.

    :param filename: The file path to the tar file.
    :return: 'gz', 'bz2' or 'xz' if the file is compressed by one of these methods, otherwise an empty string.
    """
    extension = splitext(filename)[1]
    return extension[1:].lower() if extension[1:].lower() in COMPRESS_METHODS else ''


def open_tar_file(tar_file: str, filename: str) -> IO:
    """ Open a tar file and return a IO stream to the file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :return: An IO stream.
    """
    compress_method = detect_compress_method(tar_file)
    tar = tarfile.open(tar_file, f'r:{compress_method}')
    file = tar.extractfile(filename)
    old_close = file.close

    def close():
        tar.close()
        old_close()

    file.close = close
    return file


def load_tar_json(tar_file: str, filename: str) -> Any:
    """ Load an object from a JSON file stored in a tar file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :return: The loaded object.
    """
    with open_tar_file(tar_file, filename) as file:
        if filename.lower().endswith('.gz'):
            return json.load(gzip.open(file))
        return json.load(file)


def load_tar_pickle(tar_file: str, filename: str) -> Any:
    """ Load an object from a pickle file stored in a tar file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :return: The loaded object.
    """
    with open_tar_file(tar_file, filename) as file:
        if filename.lower().endswith('.gz'):
            return pickle.load(gzip.open(file))
        return pickle.load(file)


def list_tar(tar_file: str) -> List[TarInfo]:
    """ Obtain a list with the information of each file or directory of a tar file.

    :param tar_file: The path to the tar file.
    :return: A list of TarInfo instances.
    """
    compress_method = detect_compress_method(tar_file)
    with tarfile.open(tar_file, f'r:{compress_method}') as tar:
        return tar.getmembers()


def extract_tar_file(tar_file: str, dest: str, filename: str) -> None:
    """ Extract a file inside of a tar archive.

    :param tar_file: The path to the tar file.
    :param dest: The file or folder where the tar archive should be extracted.
    :param filename: The relative path inside of the tar file to extract.
    """
    compress_method = detect_compress_method(tar_file)
    dest = join(dest, filename) if isdir(dest) else dest
    with tarfile.open(tar_file, f'r:{compress_method}') as tar:
        with tar.extractfile(filename) as reader:
            if dirname(dest) and not exists(dirname(dest)):
                makedirs(dirname(dest))
            with open(dest, 'wb') as writer:
                for chunk in reader:
                    writer.write(chunk)


def extract_tar_files(tar_file: str, dest: str, *files: str, force: bool = False, verbose: bool = False) -> None:
    """ Extract a file inside of a tar archive.

    :param tar_file: The path to the tar file.
    :param dest: The folder where the tar archive should be extracted.
    :param files: The relative path inside of the tar file to extract.
    :param force: If True, create the destination directory if it doesn't exist.
      If files is not given, then extract all the files.
    """
    if not exists(dest) and force:
        makedirs(dest)
    if not isdir(dest):
        raise ValueError(f'The destination is not an existing folder.')
    files = files if files else [file.path for file in list_tar(tar_file)]
    for file in tqdm(files, desc='Extracting files', disable=not verbose):
        extract_tar_file(tar_file, dest, file)


def extract_tar(tar_file: str, dest: str, force: bool = False, verbose: bool = False) -> None:
    """ Extract all the files inside of a tar file in the specified directory.

    :param tar_file: The tar file to extract.
    :param dest: The destination directory.
    :param force: If True, create the destination directory if it doesn't exist.
    :param verbose: If verbose, show the progress bar.
    """
    if not exists(dest) and force:
        makedirs(dest)
    if not exists(dest):
        raise FileNotFoundError(f'The folder "{dest}" does not exists. Create it or put the parameter force to True.')

    files = list_tar(tar_file)
    for file in tqdm(files, desc='Extracting files', disable=not verbose):
        if file.isdir():
            makedirs(join(dest, file.path))
        else:
            extract_tar_file(tar_file, dest, file.path)



from sys import stdout
from typing import List, Dict, TextIO

from sklearn.metrics import accuracy_score, balanced_accuracy_score, f1_score, precision_score, recall_score, \
    jaccard_score

# MEASURES ATTRIBUTES
SIMPLE_ACCURACY = 'simple_accuracy'
BALANCED_ACCURACY = 'balanced_accuracy'
MICRO_F1 = 'micro_f1'
MACRO_F1 = 'macro_f1'
WEIGHTED_F1 = 'weighted_f1'
MICRO_PRECISION = 'micro_precision'
MACRO_PRECISION = 'macro_precision'
WEIGHTED_PRECISION = 'weighted_precision'
MICRO_RECALL = 'micro_recall'
MACRO_RECALL = 'macro_recall'
WEIGHTED_RECALL = 'weighted_recall'
MICRO_JACCARD = 'micro_jaccard'
MACRO_JACCARD = 'macro_jaccard'
WEIGHTED_JACCARD = 'weighted_jaccard'
ALL_METRICS = (SIMPLE_ACCURACY, BALANCED_ACCURACY, MICRO_F1, MACRO_F1, WEIGHTED_F1, MICRO_PRECISION, MACRO_PRECISION,
               WEIGHTED_PRECISION, MICRO_RECALL, MACRO_RECALL, WEIGHTED_RECALL, MICRO_JACCARD, MACRO_JACCARD,
               WEIGHTED_JACCARD)


def metrics(trues: List[int], predictions: List[int]) -> Dict[str, float]:
    """
    Calculate the different metrics from the list of real classes and the predicted ones.
    :param trues: The real classes for each sample.
    :param predictions: The predicted classes for each sample.
    :return: A dictionary with the measure names and their respective values.
    """
    met = {SIMPLE_ACCURACY: accuracy_score(trues, predictions),
           BALANCED_ACCURACY: balanced_accuracy_score(trues, predictions),
           MICRO_F1: f1_score(trues, predictions, average='micro'),
           MACRO_F1: f1_score(trues, predictions, average='macro'),
           WEIGHTED_F1: f1_score(trues, predictions, average='weighted'),
           MICRO_PRECISION: precision_score(trues, predictions, average='micro'),
           MACRO_PRECISION: precision_score(trues, predictions, average='macro'),
           WEIGHTED_PRECISION: precision_score(trues, predictions, average='weighted'),
           MICRO_RECALL: recall_score(trues, predictions, average='micro'),
           MACRO_RECALL: recall_score(trues, predictions, average='macro'),
           WEIGHTED_RECALL: recall_score(trues, predictions, average='weighted'),
           MICRO_JACCARD: jaccard_score(trues, predictions, average='micro'),
           MACRO_JACCARD: jaccard_score(trues, predictions, average='macro'),
           WEIGHTED_JACCARD: jaccard_score(trues, predictions, average='weighted')}
    return met


def print_metrics(metrics: Dict[str, float], measures: List[str] = ALL_METRICS, file: TextIO = stdout) -> None:
    """
    Print the metrics of an evaluation.
    :param metrics: The a dictionary with the metrics to print.
    :param measures: The list to measures to print.
    :param file: The file handler where the result are printed. By default in the standard output.
    """
    _print_metric_if_show('Simple accuracy', metrics, SIMPLE_ACCURACY, measures, file)
    _print_metric_if_show('Balanced accuracy', metrics, BALANCED_ACCURACY, measures, file)
    print(file=file)
    _print_metric_if_show('Micro f-measure', metrics, MICRO_F1, measures, file)
    _print_metric_if_show('Macro f-measure', metrics, MACRO_F1, measures, file)
    _print_metric_if_show('Weighted f-measure', metrics, WEIGHTED_F1, measures, file)
    print(file=file)
    _print_metric_if_show('Micro precision', metrics, MICRO_PRECISION, measures, file)
    _print_metric_if_show('Macro precision', metrics, MACRO_PRECISION, measures, file)
    _print_metric_if_show('Weighted precision', metrics, WEIGHTED_PRECISION, measures, file)
    print(file=file)
    _print_metric_if_show('Micro recall', metrics, MICRO_RECALL, measures, file)
    _print_metric_if_show('Macro recall', metrics, MACRO_RECALL, measures, file)
    _print_metric_if_show('Weighted recall', metrics, WEIGHTED_RECALL, measures, file)
    print(file=file)
    _print_metric_if_show('Micro Jaccard', metrics, MICRO_JACCARD, measures, file)
    _print_metric_if_show('Macro Jaccard', metrics, MACRO_JACCARD, measures, file)
    _print_metric_if_show('Weighted Jaccard', metrics, WEIGHTED_JACCARD, measures, file)
    print(file=file)


def _print_metric_if_show(msg: str, metrics: Dict[str, float], metric: str, show: List[str], file: TextIO):
    """
    Print a given metric if that metric is selected.
    :param msg: The message to print with the metric.
    :param metrics: All obtained metrics.
    :param metric: The metric to print.
    :param show: The list of metrics which will be printed.
    :param file: The file handler where the result are printed. By default in the standard output.
    """
    if metric in show:
        print(f'{msg}: ', format_value(metrics[metric]), end='\t', file=file)


def format_value(value: float) -> str:
    """
    Format the metrics.
    :param value: The value to format.
    :return: The formated value.
    """
    return '{0:.2f}%'.format(value * 100)



from typing import Union, List


def head(obj: Union[dict, set], top: int = 10) -> Union[dict, set]:
    """ Obtain the first top elements of a dictionary o a set.

    :param obj: A dictionary or a set.
    :param top: The number of elements to return, by default 10.
    :return: The first top elements of the given dictionary or set.
    """
    if isinstance(obj, dict):
        return dh(obj, top)
    if isinstance(obj, set):
        return sh(obj, top)
    raise ValueError('The object must be a dictionary or a set.')


def dh(d: dict, top: int = 10) -> dict:
    """ Obtain the first top elements of a dictionary.

    :param d: The dictionary.
    :param top: The number of dictionary items to return
    :return: A dictionary with the top first items of the given dictionary.
    """
    return {key: value for i, (key, value) in enumerate(d.items()) if i < top}


def sh(d: set, top: int = 10) -> set:
    """ Obtain the first top elements of a set.

    :param d: The set.
    :param top: The number of set items to return
    :return: A set with the top first items of the given set.
    """
    return {value for i, value in enumerate(sorted(d)) if i < top}


def list_union(*lists: list) -> list:
    """ Create a new list with the union of the lists without repeated elements.

    :param lists: The list to union.
    :return: The union of all lists respecting the element order.
    """
    result = []
    for l in lists:
        for e in l:
            if e not in result:
                result.append(e)
    return result



from logging import getLogger
from os.path import exists
from threading import Thread, Event
from typing import Callable, List

try:
    import git
except ImportError:
    raise ModuleNotFoundError('git is required for this command. '
                              'Please, install it with:\n\npip install GitPython~=3.1.11')

logger = getLogger(__name__)


class GitMonitor(object):
    """ A Git monitor to detect changes in a branch. """
    def __init__(self, func: Callable, folder: str, repository: str, branch: str = None, force: bool = False,
                 interval: int = 0) -> None:
        """ Monitor a chatbot data repository.
        :param func: The function to call when the repository changes or
          the first time if the force argument is activated.
        :param folder: The path to the local git repository.
        :param repository: The url to the git repository.
        :param branch: The git branch to monitor. By default, the current branch is used.
        :param interval: The time in seconds to wait before checking for repository changes.
        :param force: If force the update model independently if there are changes in the repository or not.
          If this argument is set, then the model is not copied to the S3 bucket.
        """
        self.func = func
        self.folder = folder
        self.repository = repository
        self.branch = branch if branch else 'master'
        self.force = force
        self.interval = interval
        self.updated = True
        self.__stop = False
        self.__thread = None
        self.__event = Event()

    def monitor(self, update: bool = True, **kwargs) -> None:
        """ Start the monitor.
        :param update: If the local repository is updated after to check changes.
        :param kwargs: Extra arguments to pass to the function.
        """
        force = self.force
        if exists(self.folder):
            repo = git.Repo(self.folder)
        else:
            force = True
            repo = git.Repo.clone_from(self.repository, self.folder)
        if not self.branch:
            self.branch = repo.active_branch.name
        elif repo.active_branch.name != self.branch:
            repo.git.checkout(self.branch)
        updated = True
        while not self.__stop:
            changes = self.__changes(repo)
            if changes or force:
                if update:
                    repo.git.pull()
                    updated = True
                self.func(*changes, **kwargs)
            if not self.interval:
                break
            if updated:
                logger.info(f'Waiting {self.interval}s for changes in the configuration. ')
            self.__event.wait(self.interval)
            force = False
            updated = False
        self.__thread = None
        self.__stop = False

    def __changes(self, repo: git.Repo) -> List[str]:
        """ Get the changes between local and remote repository for the set branch.

        :param repo: The Git repository.
        :return: A list with the changed file names.
        """
        origin = repo.remotes.origin
        fetch = [info for info in origin.fetch() if info.name == f'origin/{self.branch}'][0]
        diff = fetch.commit.diff()
        return [d.a_path for d in diff]

    def start(self, update: bool = True, **kwargs) -> None:
        """ Start the monitor as a thread.

        :param update: If the local repository is updated after to check changes.
        """
        self.__thread = Thread(target=self.monitor, args=(update,), kwargs=kwargs)
        self.__thread.start()

    def stop(self):
        """ Stop the monitor. """
        self.__stop = True
        self.__event.set()



from subprocess import Popen, PIPE
from typing import Tuple, List, Union
import re


def split_arg_string(command: str) -> List[str]:
    """ Given an argument string this attempts to split it into small command parts.

    :param command: The command to split.
    :return: A list with each  separated command parts.
    """

    rv = []
    for match in re.finditer(r"('([^'\\]*(?:\\.[^'\\]*)*)'"
                             r'|"([^"\\]*(?:\\.[^"\\]*)*)"'
                             r'|\S+)\s*', command, re.S):
        arg = match.group().strip()
        if arg[:1] == arg[-1:] and arg[:1] in '"\'':
            arg = arg[1:-1].encode('ascii', 'backslashreplace').decode('unicode-escape')
        try:
            arg = type(command)(arg)
        except UnicodeError:
            pass
        rv.append(arg)
    return rv


def execute_command(command: Union[List[str], str], input_text: str = '') -> Tuple[str, str]:
    """ Execute an external command easily.

    :param command: The command.
    :param input_text: The text of the standard input.
    :return: A tuple with the standard and error outputs.
    """
    command = split_arg_string(command) if isinstance(command, str) else command
    process = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)

    process.stdin.write(f'{input_text}\n')

    try:
        return process.communicate()
    finally:
        process.stdin.close()



import unittest

from mysutils.file import exist_files, not_exist_files, are_dir, not_are_dir


class FileTestCase(unittest.TestCase):
    """ Class for especial file tests. """

    def assertExists(self, *files: str) -> None:
        """ Check if all of the files exist.

        :param files: The list of file paths to check.
        """
        self.assertTrue(exist_files(*files))

    def assertNotExists(self, *files: str) -> None:
        """ Check if any of the files exist.

        :param files: The list of file paths to check.
        """
        self.assertTrue(not_exist_files(*files))

    def assertAreDir(self, *files: str) -> None:
        """ Check if any of the files exist.

        :param files: The list of file paths to check.
        """
        self.assertTrue(are_dir(*files))

    def assertNotAreDir(self, *files: str) -> None:
        """ Check if any of the files exist.

        :param files: The list of file paths to check.
        """
        self.assertTrue(not_are_dir(*files))


def main():
    unittest.main()



from typing import Tuple, Any, List


def parse_config(config: dict, params: List[Tuple[str, bool, Any]], double_check: bool = True) -> Tuple:
    """  Check the configuration from a list of parameter definition.

    :param config: The configuration to check.
    :param params: The list of parameter definitions. This list is a tuple with the name of the parameter,
       if the parameter is or not required, and it default value if it is not required. For example, the tuple
       ('server_host', False, 'http://0.0.0.0') defines that the parameter name is 'server_host', that it is not
       required, and if it is not given, then the value is 'http://0.0.0.0'.
    :param double_check: With True, check if all the required configuration parameters are present, and if there are
       any extra parameter which are not present in the parameter definition list. If False, only check the first one:
       if all the required parameters are present.
    :return: The parameter values.
    """
    # Check if there are parameters in the configuration file that are not defined
    if double_check:
        for key in config:
            if key not in [p[0] for p in params]:
                raise ValueError(f'The parameter "{key}" in the configuration file is not a valid parameter.')
    # Check if all required parameters are present in the configuration file
    for key in {p[0] for p in params if p[1]}:
        if key not in config:
            raise ValueError(f'The parameter "{key}" is not defined in the configuration file and it is mandatory.')
    # Extract the parameter values
    values = []
    for p in params:
        values.append(config[p[0]] if p[0] in config else p[2])
    return tuple(values)



from threading import Lock
from typing import Callable


def synchronized(method: Callable) -> Callable:
    """ Function decorator to synchronize a method.

    :param method: The method.
    :return: The synchronized method.
    """
    outer_lock = Lock()
    lock_name = "__" + method.__name__ + "_lock" + "__"

    def sync_method(self, *args, **kws):
        with outer_lock:
            if not hasattr(self, lock_name):
                setattr(self, lock_name, Lock())
            lock = getattr(self, lock_name)
            with lock:
                return method(self, *args, **kws)

    return sync_method
